import logging
import os

from PyQt6 import uic
# will change these to specific imports once code is more final
from PyQt6.QtCore import *
from PyQt6.QtGui import *
from PyQt6.QtWidgets import *

from mpfmonitor._version import __version__, __bcp_version__
from mpfmonitor.core.playfield import Shape


class InspectorWindow(QWidget):

    def __init__(self, mpfmon):
        self.mpfmon = mpfmon
        super().__init__()
        self.ui = None

        self.log = logging.getLogger('Core')

        self.draw_ui()
        self.attach_signals()

        self.enable_non_default_widgets(enabled=False)

        self.last_pf_widget = None

    def draw_ui(self):
        # Load ui file from ./ui/
        ui_path = os.path.join(os.path.dirname(__file__), "ui", "inspector.ui")
        self.ui = uic.loadUi(ui_path, self)

        self.ui.setWindowTitle('Inspector')

        self.ui.move(self.mpfmon.local_settings.value('windows/inspector/pos',
                                                   QPoint(1100, 465)))
        self.ui.resize(self.mpfmon.local_settings.value('windows/inspector/size',
                                                     QSize(300, 340)))

        mpf_monitor_version = "MPF Monitor Version: {}".format(__version__)
        self.ui.mpf_monitor_version.setText(mpf_monitor_version)
        bcp_required_version = "BCP Version Required: {} or greater".format(__bcp_version__)
        self.ui.bcp_required_version.setText(bcp_required_version)

    def attach_signals(self):
        self.attach_inspector_tab_signals()
        self.attach_monitor_tab_signals()

    def attach_inspector_tab_signals(self):
        self.ui.toggle_inspector_button.clicked.connect(self.toggle_inspector_mode)

        self.ui.shape_combo_box.currentIndexChanged.connect(self.shape_combobox_changed)
        self.ui.rotationDial.valueChanged.connect(self.dial_changed)

        self.ui.size_slider.valueChanged.connect(self.slider_drag)  # Doesn't save value, just for live preview
        self.ui.size_slider.sliderReleased.connect(self.slider_changed)  # Saves value on release
        self.ui.size_spinbox.valueChanged.connect(self.spinbox_changed)

        self.ui.reset_to_defaults_button.clicked.connect(self.reset_defaults_last_device)
        self.ui.delete_last_device_button.clicked.connect(self.delete_last_device)

    def attach_monitor_tab_signals(self):
        self.ui.toggle_device_win_button.setChecked(self.mpfmon.toggle_device_window_action.isChecked())
        self.ui.toggle_device_win_button.stateChanged.connect(self.mpfmon.toggle_device_window)

        self.ui.toggle_event_win_button.setChecked(self.mpfmon.toggle_event_window_action.isChecked())
        self.ui.toggle_event_win_button.stateChanged.connect(self.mpfmon.toggle_event_window)

        self.ui.toggle_pf_win_button.setChecked(self.mpfmon.toggle_pf_window_action.isChecked())
        self.ui.toggle_pf_win_button.stateChanged.connect(self.mpfmon.toggle_pf_window)

        self.ui.toggle_mode_win_button.setChecked(self.mpfmon.toggle_mode_window_action.isChecked())
        self.ui.toggle_mode_win_button.stateChanged.connect(self.mpfmon.toggle_mode_window)

        self.ui.toggle_variables_win_button.setChecked(self.mpfmon.toggle_variables_window_action.isChecked())
        self.ui.toggle_variables_win_button.stateChanged.connect(self.mpfmon.toggle_variables_window)

        self.ui.exit_on_close_button.setChecked(self.mpfmon.get_local_settings_bool('settings/exit-on-close'))
        self.ui.exit_on_close_button.stateChanged.connect(self.mpfmon.toggle_exit_on_close)

    def toggle_inspector_mode(self):
        inspector_enabled = not self.mpfmon.inspector_enabled
        if self.registered_inspector_cb:
            self.set_inspector_val_cb(inspector_enabled)
            if inspector_enabled:
                self.log.debug('Inspector mode toggled ON')
            else:
                self.log.debug('Inspector mode toggled OFF')
                self.enable_non_default_widgets(enabled=False)
                self.clear_last_selected_device()

    def register_set_inspector_val_cb(self, cb):
        self.registered_inspector_cb = True
        self.set_inspector_val_cb = cb

    def update_last_selected(self, pf_widget=None):
        if pf_widget is not None:
            self.enable_non_default_widgets(enabled=True)

            self.last_pf_widget = pf_widget

            # Update the label to show name of last selected
            text = '"' + str(self.last_pf_widget.name) + '" Size:'
            self.ui.device_group_box.setTitle(text)

            # Update the size slider and spinbox
            self.ui.size_slider.setValue(int(self.last_pf_widget.size * 100))
            self.ui.size_spinbox.setValue(self.last_pf_widget.size)

            # Update the shape combo box
            self.ui.shape_combo_box.setCurrentIndex(self.last_pf_widget.shape_type.value)

            # Update the rotation dial
            rotation = int(self.last_pf_widget.angle / 10) + 18
            self.ui.rotationDial.setValue(rotation)

    def slider_drag(self):
        # For live preview
        new_size = self.ui.size_slider.value() / 100  # convert from int to float
        self.update_last_device(new_size=new_size, save=False)

    def slider_changed(self):
        new_size = self.ui.size_slider.value() / 100  # convert from int to float
        # Update spinbox value
        self.ui.size_spinbox.setValue(new_size)

        # Don't need to call resize_last_device because updating the spinbox takes care of it
        # self.resize_last_device(new_size=new_size)

    def spinbox_changed(self):
        new_size = self.ui.size_spinbox.value()
        # Update slider value
        self.ui.size_slider.setValue(int(new_size*100))

        self.update_last_device(new_size=new_size)

    def dial_changed(self):
        rot_value = self.ui.rotationDial.value() * 10
        # Offset the dial by 180
        rot_value = (rot_value - 180) % 360
        # self.rotate_last_device(rotation=rot_value, save=False)
        self.update_last_device(rotation=rot_value, save=True)

    def shape_combobox_changed(self):
        shape_index = self.ui.shape_combo_box.currentIndex()
        self.update_last_device(shape=Shape(shape_index), save=True)

    def clear_last_selected_device(self):
        # Must be called AFTER spinbox valueChanged is set. Otherwise slider will not follow

        # self.last_selected_label.setText("Default Device Size:")
        self.ui.device_group_box.setTitle("Default Device:")
        self.last_pf_widget = None
        self.ui.size_spinbox.setValue(self.mpfmon.pf_device_size)  # Reset the value to the stored default.
        self.enable_non_default_widgets(enabled=False)


    def enable_non_default_widgets(self, enabled=False):
        if self.ui is not None:
            self.ui.rotationDial.setEnabled(enabled)
            self.ui.shape_combo_box.setEnabled(enabled)


    def update_last_device(self, new_size=None, rotation=None, shape=None, save=True):
        # Check that there is a last widget
        if self.last_pf_widget is not None:

            update_and_resize = False

            if new_size is not None:
                new_size = round(new_size, 3)

                self.last_pf_widget.set_size(new_size)
                update_and_resize = True

            if rotation is not None:
                self.last_pf_widget.set_rotation(rotation)
                update_and_resize = True

            if shape is not None:
                self.last_pf_widget.set_shape_type(shape_type=shape)
                update_and_resize = True

            if update_and_resize:
                self.last_pf_widget.update_pos(save=save)
                self.mpfmon.view.resizeEvent()

        else:
            if new_size is not None:
                self.mpfmon.pf_device_size = new_size
                self.mpfmon.config["device_size"] = new_size

            if save:
                self.resize_all_devices()  # Apply new sizes to all devices without default sizes
                self.mpfmon.view.resizeEvent()  # Re draw the playfield
                self.mpfmon.save_config()  # Save the config with new default to disk


    def delete_last_device(self):
        if self.last_pf_widget is not None:
            self.last_pf_widget.destroy()
            self.clear_last_selected_device()
        else:
            self.log.info("No device selected to delete")

    def reset_defaults_last_device(self):
        if self.last_pf_widget is not None:

            # Redraw the device and save changes
            default_size = self.mpfmon.pf_device_size
            self.update_last_device(new_size=default_size, shape=Shape.DEFAULT,
                                    rotation=0, save=True)

            self.ui.size_spinbox.setValue(default_size)
            self.ui.rotationDial.setValue(18)
            self.ui.shape_combo_box.setCurrentIndex(0)


            # Update the device info and clear saved size data
            self.last_pf_widget.resize_to_default(force=True)


            # Redraw the device
        else:
            self.ui.size_spinbox.setValue(0.07)
            self.log.info("No device selected to resize")

    def resize_all_devices(self):
        for i in self.mpfmon.scene.items():
            try:
                i.resize_to_default()
            except AttributeError as e:
                # Can't resize object. That's ok.
                pass

    def register_last_selected_cb(self):
        self.mpfmon.inspector_window_last_selected_cb = self.update_last_selected


    def closeEvent(self, event):
        self.mpfmon.write_local_settings()
        event.accept()
        self.mpfmon.check_if_quit()



import unittest

from mpfmonitor.core.inspector import *
from unittest.mock import MagicMock


class TestableInspectorNoGUI(InspectorWindow):
    def __init__(self, mpfmon_mock=None, logger=False):
        if mpfmon_mock is not None:
            self.mpfmon = mpfmon_mock

        # super().__init__()
        self.ui = None

        # Call logger=True if "RuntimeError: super-class __init__()" starts failing tests.
        if logger:
            self.log = logging.getLogger('Core')

        # self.draw_ui()
        # self.attach_signals()


class InspectorMode(unittest.TestCase):
    def test_toggle_inspector_mode_on(self):
        mock_mpfmon = MagicMock()
        mock_mpfmon.inspector_enabled = False
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon, logger=True)

        # Register the callback to set the inspector value as a mock
        inspector.register_set_inspector_val_cb(MagicMock())

        inspector.toggle_inspector_mode()

        # Test that the previously registered mock is called.
        inspector.set_inspector_val_cb.assert_called_once_with(True)

    def test_toggle_inspector_mode_off(self):
        mock_mpfmon = MagicMock()
        mock_mpfmon.inspector_enabled = True
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon, logger=True)

        # Register the callback to set the inspector value as a mock
        inspector.register_set_inspector_val_cb(MagicMock())

        # clear_last_selected_device should be called as a result of toggling on -> off
        inspector.clear_last_selected_device = MagicMock()

        inspector.toggle_inspector_mode()

        # Test that the previously registered mock is called.
        inspector.set_inspector_val_cb.assert_called_once_with(False)

        # Test clear_last_selected_device was actually called
        inspector.clear_last_selected_device.assert_called_once()

    def test_cb_register(self):
        inspector = TestableInspectorNoGUI()
        inspector.registered_inspector_cb = False

        callback = MagicMock()
        inspector.register_set_inspector_val_cb(cb=callback)

        # Check that the mocked callback is registered properly
        inspector.set_inspector_val_cb(True)

        self.assertTrue(inspector.registered_inspector_cb)
        inspector.set_inspector_val_cb.assert_called_once_with(True)

class InspectorDeviceManipulation(unittest.TestCase):

    def test_update_last_selected(self):
        inspector = TestableInspectorNoGUI()
        # Mock ui to check that the ui is updated
        inspector.ui = MagicMock()

        # Mock widget to pass into the pf_widget parameter
        mock_widget = MagicMock()

        # Set the mock widget's name and size.
        mock_widget.name.__str__.return_value = 'LastName'
        # Return value doesn't quite work here. Just load in a float.
        widget_size = float(0.10)
        mock_widget.size = widget_size

        inspector.update_last_selected(pf_widget=mock_widget)

        # Check that the name is called.
        mock_widget.name.__str__.assert_called_once()

        inspector.ui.device_group_box.setTitle.assert_called_once_with('"LastName" Size:')
        inspector.ui.size_slider.setValue.assert_called_once_with(widget_size * 100)
        inspector.ui.size_spinbox.setValue.assert_called_once_with(widget_size)

    def test_delete_last_device(self):
        inspector = TestableInspectorNoGUI()

        inspector.last_pf_widget = MagicMock()
        inspector.clear_last_selected_device = MagicMock()

        inspector.delete_last_device()

        inspector.last_pf_widget.destroy.assert_called_once()
        inspector.clear_last_selected_device.assert_called_once()


class InspectorDeviceResizing(unittest.TestCase):

    def test_resize_default_device_default_no_save(self):
        mock_mpfmon = MagicMock()
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon)

        size = float(0.07)

        inspector.last_pf_widget = None
        inspector.update_last_device(new_size=size, save=False)

        # self.assertEqual(mock_mpfmon.pf_device_size, size)

    def test_resize_default_device_default_save(self):
        mock_mpfmon = MagicMock()
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon)

        size = float(0.07)

        inspector.last_pf_widget = None
        inspector.resize_all_devices = MagicMock()
        inspector.update_last_device(new_size=size, save=True)

        # self.assertEqual(mock_mpfmon.pf_device_size, size)
        inspector.resize_all_devices.assert_called_once()
        mock_mpfmon.view.resizeEvent.assert_called_once()  # Re draw the playfiled
        mock_mpfmon.save_config.assert_called_once()  # Save the config with new default to disk

    def test_resize_last_device_default_no_save(self):
        mock_mpfmon = MagicMock()
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon)

        size = float(0.07)

        inspector.last_pf_widget = MagicMock()
        inspector.update_last_device(new_size=size, save=False)

        inspector.last_pf_widget.set_size.assert_called_once_with(size)
        inspector.last_pf_widget.update_pos.assert_called_once_with(save=False)

        mock_mpfmon.view.resizeEvent.assert_called_once()  # Re draw the playfield

    def test_resize_last_device_default_save(self):
        mock_mpfmon = MagicMock()
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon)

        size = float(0.07)

        inspector.last_pf_widget = MagicMock()
        inspector.update_last_device(new_size=size, save=True)

        inspector.last_pf_widget.set_size.assert_called_once_with(size)
        inspector.last_pf_widget.update_pos.assert_called_once_with(save=True)

        mock_mpfmon.view.resizeEvent.assert_called_once()  # Re draw the playfield


class InspectorDeviceRotation(unittest.TestCase):

    def test_rotate_device_without_save(self):
        mock_mpfmon = MagicMock()
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon)

        rotation = 90

        inspector.last_pf_widget = MagicMock()
        inspector.update_last_device(rotation=rotation, save=False)
        inspector.last_pf_widget.set_rotation.assert_called_once_with(rotation)
        inspector.last_pf_widget.update_pos.assert_called_once_with(save=False)

    def test_rotate_device_with_save(self):
        mock_mpfmon = MagicMock()
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon)

        rotation = 90

        inspector.last_pf_widget = MagicMock()
        inspector.update_last_device(rotation=rotation, save=True)
        inspector.last_pf_widget.set_rotation.assert_called_once_with(rotation)
        inspector.last_pf_widget.update_pos.assert_called_once_with(save=True)


class InspectorDeviceShape(unittest.TestCase):
    from mpfmonitor.core.playfield import Shape

    def test_device_shape_without_save(self):
        mock_mpfmon = MagicMock()
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon)

        shape = Shape.TRIANGLE

        inspector.last_pf_widget = MagicMock()
        inspector.update_last_device(shape=shape, save=False)
        inspector.last_pf_widget.set_shape_type.assert_called_once_with(shape_type=shape)
        inspector.last_pf_widget.update_pos.assert_called_once_with(save=False)

    def test_device_shape_with_save(self):
        mock_mpfmon = MagicMock()
        inspector = TestableInspectorNoGUI(mpfmon_mock=mock_mpfmon)

        shape = Shape.TRIANGLE

        inspector.last_pf_widget = MagicMock()
        inspector.update_last_device(shape=shape, save=True)
        inspector.last_pf_widget.set_shape_type.assert_called_once_with(shape_type=shape)
        inspector.last_pf_widget.update_pos.assert_called_once_with(save=True)


if __name__ == '__main__':
    unittest.main()



"""BCP Server interface for the MPF Media Controller"""

import logging
import queue
import socket
import threading
import os
import time

import select

from datetime import datetime
import math

import mpf.core.bcp.bcp_socket_client as bcp
from PyQt6.QtCore import QTimer


class BCPClient(object):

    def __init__(self, mpfmon, receiving_queue, sending_queue,
                 interface='localhost', port=5051, simulate=False, cache=False):

        self.mpfmon = mpfmon
        self.log = logging.getLogger('BCP Client')
        self.interface = interface
        self.port = port
        self.receive_queue = receiving_queue
        self.sending_queue = sending_queue
        self.connected = False
        self.socket = None
        self.sending_thread = None
        self.receive_thread = None
        self.connection_thread_running = False
        self.done = False
        self.last_time = datetime.now()

        self.simulate = simulate
        self.caching_enabled = cache

        try:
            self.cache_file_location = os.path.join(self.mpfmon.machine_path, "monitor", "cache.txt")
        except FileNotFoundError:
            self.simulate = False
            self.caching_enabled = False

        self.mpfmon.log.info('Looking for MPF at %s:%s', self.interface, self.port)

        self.simulator_timer = QTimer(self.mpfmon.device_window)

        self.simulator_messages = []
        self.simulator_msg_timer = []
        self.enable_simulator(enable=self.simulate)

    def start_connect_thread(self):
        """Starts the connection process in a separate thread."""
        if not self.connection_thread_running:
            self.connection_thread_running = True
            self.connection_thread = threading.Thread(target=self.connection_thread, daemon=True)
            self.connection_thread.start()

    def connection_thread(self):
        """The main loop for the BCP client running in a separate thread."""
        while not self.mpfmon.thread_stopper.is_set():
            if not self.connected:
                self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                try:
                    self.socket.connect((self.interface, int(self.port)))
                    self.socket.settimeout(0.1) # Set a small timeout for non-blocking operations
                    self.connected = True
                    self.log.info("Connected to MPF.")
                except socket.error as e:
                    self.connected = False
                    time.sleep(1)
                    continue

                if self.connected:
                    if self.create_socket_threads():
                        self.start_monitoring()
            else:
                time.sleep(1)


    def register_timer(self):
        if self.simulate:
            self.simulator_init()
            self.simulator_timer.setInterval(100)
            self.simulator_timer.timeout.connect(self.simulate_received)
            self.simulator_timer.start()
        else:
            self.simulator_timer.stop()
            self.start_connect_thread()

    def enable_simulator(self, enable=True):
        if enable:
            self.simulate = True
            if self.caching_enabled:
                try:
                    self.cache_file = open(self.cache_file_location, "r")
                except FileNotFoundError:
                    self.log.warn("Caching enabled but no cache file found.")
                    self.simulate = False
                    self.caching_enabled = False
                    self.enable_simulator(False)
        else:
            self.simulate = False
            if self.caching_enabled:
                self.cache_file = open(self.cache_file_location, "w")

        self.start_time = datetime.now()
        self.register_timer()

    def start_monitoring(self):
        self.sending_queue.put('monitor_start?category=devices')
        self.sending_queue.put('monitor_start?category=events')
        self.sending_queue.put('monitor_start?category=modes')
        self.sending_queue.put('monitor_start?category=machine_vars')
        self.sending_queue.put('monitor_start?category=player_vars')

    def create_socket_threads(self):
        """Creates and starts the sending and receiving threads for the BCP
        socket.
        Returns:
            True if the socket exists and the threads were started. False if
            not.
        """

        if self.socket:

            self.receive_thread = threading.Thread(target=self.receive_loop)
            # self.receive_thread.daemon = True
            self.receive_thread.start()

            self.sending_thread = threading.Thread(target=self.sending_loop)
            # self.sending_thread.daemon = True
            self.sending_thread.start()

            return True

        else:
            return False

    def receive_loop(self):
        """The socket thread's run loop."""
        socket_chars = b''
        while self.connected and not self.mpfmon.thread_stopper.is_set():
            try:
                ready = select.select([self.socket], [], [], 1)
                if ready[0]:
                    data_read = self.socket.recv(8192)
                    if data_read:
                        socket_chars += data_read
                        commands = socket_chars.split(b"\n")

                        # keep last incomplete command
                        socket_chars = commands.pop()

                        # process all complete commands
                        for cmd in commands:
                            if cmd:
                                self.process_received_message(cmd.decode())
                    else:
                        # no bytes -> socket closed
                        break

            except socket.timeout:
                pass

            except OSError:
                break

        self.connected = False

    def disconnect(self):
        if not self.connected:
            self.log.info("Disconnecting from BCP")
            self.sending_queue.put('goodbye', None)

    def close(self):
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
            self.socket.close()

        except (OSError, AttributeError):
            pass

        if self.caching_enabled:
            self.cache_file.close()

        self.socket = None
        self.connected = False

        with self.receive_queue.mutex:
            self.receive_queue.queue.clear()

        with self.sending_queue.mutex:
            self.sending_queue.queue.clear()

    def sending_loop(self):
        while self.connected and not self.mpfmon.thread_stopper.is_set():
            try:
                msg = self.sending_queue.get(block=True, timeout=1)
            except queue.Empty:
                if self.mpfmon.thread_stopper.is_set():
                    return

                else:
                    continue

            self.socket.sendall(('{}\n'.format(msg)).encode('utf-8'))

        self.connected = False

    def process_received_message(self, message):
        """Puts a received BCP message into the receiving queue.

        Args:
            message: The incoming BCP message

        """
        self.log.debug('Received "%s"', message)
        if self.caching_enabled and not self.simulate:
            time = datetime.now() - self.last_time
            message_tmr = math.floor(time.microseconds / 1000)
            self.last_time = datetime.now()
            self.cache_file.write(str(message_tmr) + "," + message + "\n")

        try:
            cmd, kwargs = bcp.decode_command_string(message)
            self.receive_queue.put((cmd, kwargs))
        except ValueError:
            self.log.error("DECODE BCP ERROR. Message: %s", message)
            raise

    def send(self, bcp_command, **kwargs):
            self.sending_queue.put(bcp.encode_command_string(bcp_command,
                                                             **kwargs))

    def simulator_init(self):
        if self.caching_enabled:
            self.simulator_msg_timer.append(int(0))
            for message in self.cache_file:
                message_tmr = int(message.split(',', 1)[0])
                message_str = message.split(',', 1)[1]

                self.simulator_msg_timer.append(message_tmr)
                self.simulator_messages.append(message_str)
        else:
            messages = [
                'device?json={"type": "switch", "name": "s_start", "changes": false, "state": {"state": 0, "recycle_jitter_count": 0}}',
                'device?json={"type": "switch", "name": "s_trough_1", "changes": false, "state": {"state": 1, "recycle_jitter_count": 0}}',
                'device?json={"type": "light", "name": "l_shoot_again", "changes": ["color", [255, 255, 255], [0, 0, 0]], "state": {"color": [0, 0, 0]}}',
                'device?json={"type": "light", "name": "l_ball_save", "changes": ["color", [0, 0, 0], [255, 255, 255]], "state": {"color": [255, 255, 255]}}',
            ]
            self.simulator_messages = messages
            self.simulator_msg_timer = [100, 100, 100, 100]

    def simulate_received(self):
        if len(self.simulator_messages) > 0:
            next_message = self.simulator_messages.pop(0)
            timer = self.simulator_msg_timer.pop(0)
            self.simulator_timer.setInterval(timer)
            self.process_received_message(next_message)
        else:
            self.simulator_timer.stop()
            self.log.info("End of cached file reached.")



import unittest
from mpfmonitor.core.playfield import *
from unittest.mock import MagicMock

class TestablePfWidgetNonDrawn(PfWidget):
    def __init__(self, mpfmon_mock=None):

        if mpfmon_mock is not None:
            self.mpfmon = mpfmon_mock

    """
    __init__ of PfWidget:


    def __init__(self, mpfmon, widget, device_type, device_name, x, y,
                 size=None, rotation=0, shape=Shape.DEFAULT, save=True):
        super().__init__()

        self.widget = widget
        self.mpfmon = mpfmon
        self.name = device_name
        self.move_in_progress = True
        self.device_type = device_type
        self.set_size(size=size)
        self.shape_type = shape
        self.angle = rotation

        self.setToolTip('{}: {}'.format(self.device_type, self.name))
        self.setAcceptedMouseButtons(Qt.MouseButton.LeftButton | Qt.MouseButton.RightButton)
        self.setPos(x, y)
        self.update_pos(save)
        self.click_start = 0
        self.release_switch = False

        self.log = logging.getLogger('Core')

        old_widget_exists = widget.set_change_callback(self.notify)

        if old_widget_exists:
            self.log.debug("Previous widget exists.")
            old_widget_exists(destroy=True)

        """


class TestPfWidgetParameters(unittest.TestCase):

    def setUp(self):
        self.widget = TestablePfWidgetNonDrawn()

    def test_shape_set_valid(self):
        shape_to_be_set = Shape.TRIANGLE
        self.widget.set_shape_type(shape_to_be_set)

        self.assertEqual(self.widget.shape_type, shape_to_be_set)

    def test_shape_set_invalid(self):
        widget = TestablePfWidgetNonDrawn()

        shape_to_be_set = "Not_A_Shape"
        self.widget.set_shape_type(shape_to_be_set)

        self.assertEqual(self.widget.shape_type, Shape.DEFAULT)

    def test_rotation_set_valid(self):
        rotation_to_be_set = 42
        self.widget.set_rotation(rotation_to_be_set)

        self.assertEqual(self.widget.angle, rotation_to_be_set)

    def test_rotation_set_invalid(self):
        rotation_to_be_set = 451
        self.widget.set_rotation(rotation_to_be_set)

        expected_angle = rotation_to_be_set % 360

        self.assertEqual(self.widget.angle, expected_angle)

    def test_size_set_default(self):
        self.widget.mpfmon = MagicMock()
        default_size = 0.07
        scene_width = 1.00

        self.widget.mpfmon.pf_device_size = default_size
        self.widget.mpfmon.scene.width.return_value = scene_width

        self.widget.set_size()

        self.assertEqual(self.widget.size, default_size)
        self.assertEqual(self.widget.device_size, default_size * scene_width)

    def test_size_set_valid(self):
        self.widget.mpfmon = MagicMock()
        scene_width = 1.00

        self.widget.mpfmon.scene.width.return_value = scene_width

        size_to_be_set = 0.07

        self.widget.set_size(size=size_to_be_set)

        self.assertEqual(self.widget.size, size_to_be_set)
        self.assertEqual(self.widget.device_size, size_to_be_set * scene_width)


class TestPfWidgetResizeToDefault(unittest.TestCase):

    def setUp(self):
        self.mock_mpfmon = MagicMock()

        self.widget = TestablePfWidgetNonDrawn(mpfmon_mock=self.mock_mpfmon)

        self.widget.device_type = MagicMock()
        self.widget.name = MagicMock()
        self.widget.set_size = MagicMock()
        self.widget.update_pos = MagicMock()

        self.config = MagicMock()
        self.mock_mpfmon.config[self.widget.device_type].get.return_value = self.config
        self.config.get.return_value = None

        """
        def resize_to_default(self, force=False):
        device_config = self.mpfmon.config[self.device_type].get(self.name, None)

        if force:
            device_config.pop('size', None) # Delete saved size info, None is incase key doesn't exist (popped twice)

        device_size = device_config.get('size', None)

        if device_size is not None:
            # Do not change the size if it's already set
            pass
        elif device_config is not None:
            self.set_size()

        self.update_pos(save=False)  # Do not save at this point. Let it be saved elsewhere. This reduces writes."""

    def test_size_resize_to_default(self):
        self.widget.resize_to_default()

        self.mock_mpfmon.config[self.widget.device_type].get.assert_called_once_with(self.widget.name, None)
        self.widget.set_size.assert_called_once()
        self.widget.update_pos.assert_called_once_with(save=False)

    def test_size_resize_to_default_with_force(self):
        self.widget.resize_to_default(force=True)

        self.mock_mpfmon.config[self.widget.device_type].get.assert_called_once_with(self.widget.name, None)
        self.widget.set_size.assert_called_once()
        self.widget.update_pos.assert_called_once_with(save=False)
        self.config.pop.assert_called_once_with('size', None)


class TestPfWidgetColorFuncs(unittest.TestCase):

    def setUp(self):
        self.widget = TestablePfWidgetNonDrawn()

    def test_color_gamma(self):
        color_in = [0, 128, 255]
        expected_color_out = [0, 203, 255]  # Manually calculated 128 -> 203

        mock_widget = DeviceNode()
        mock_widget.setData({"color": color_in})
        mock_widget.setType('light')
        color_out = mock_widget._calculate_color_gamma_correction(color=color_in)

        self.assertEqual(color_out, expected_color_out, 'Gamma does not match expected value')

    def test_colored_brush_light(self):
        color_in = [0, 128, 255]
        expected_color_out = [0, 203, 255]  # Manually calculated 128 -> 203
        device_type = 'light'
        mock_widget = DeviceNode()
        mock_widget.setData({"color": color_in})
        mock_widget.setType(device_type)

        expected_q_brush_out = QBrush(QColor(*expected_color_out), Qt.BrushStyle.SolidPattern)
        q_brush_out = mock_widget.get_colored_brush()

        self.assertEqual(q_brush_out, expected_q_brush_out, 'Brush is not returning correct value')

    def test_colored_brush_switch_off(self):
        device_type = 'switch'
        expected_color_out = [0, 0, 0]
        mock_widget = DeviceNode()
        mock_widget.setData({'state': False})
        mock_widget.setType(device_type)

        expected_q_brush_out = QBrush(QColor(*expected_color_out), Qt.BrushStyle.SolidPattern)
        q_brush_out = mock_widget.get_colored_brush()

        self.assertEqual(q_brush_out, expected_q_brush_out, 'Brush is not returning correct value')

    def test_colored_brush_switch_on(self):
        device_type = 'switch'
        expected_color_out = [0, 255, 0]
        mock_widget = DeviceNode()
        mock_widget.setData({'state': True})
        mock_widget.setType(device_type)

        expected_q_brush_out = QBrush(QColor(*expected_color_out), Qt.BrushStyle.SolidPattern)
        q_brush_out = mock_widget.get_colored_brush()

        self.assertEqual(q_brush_out, expected_q_brush_out, 'Brush is not returning correct value')


class TestPfWidgetGetAndDestroy(unittest.TestCase):

    def setUp(self):
        self.widget = TestablePfWidgetNonDrawn(mpfmon_mock=MagicMock())

    def test_delete_from_config(self):
        device_type = MagicMock()
        self.widget.device_type = device_type

        name = "delete_test"
        self.widget.name = name

        self.widget.delete_from_config()

        self.widget.mpfmon.config[device_type].pop.assert_called_once_with(name)
        self.widget.mpfmon.save_config.assert_called_once()

    def test_send_to_inspector_window(self):
        self.widget.send_to_inspector_window()
        self.widget.mpfmon.inspector_window_last_selected_cb.assert_called_once_with(pf_widget=self.widget)



from PyQt6.QtCore import *
from PyQt6.QtGui import *
from PyQt6.QtWidgets import *
from PyQt6 import uic

import os


class EventWindow(QWidget):

    def __init__(self, mpfmon):
        self.mpfmon = mpfmon
        self.filtered_model = None
        super().__init__()
        self.ui = None
        self.model = None
        self.draw_ui()
        self.attach_model()
        self.attach_signals()

        self.already_hidden = False
        self.added_index = 0

    def draw_ui(self):
        # Load ui file from ./ui/
        ui_path = os.path.join(os.path.dirname(__file__), "ui", "searchable_table.ui")
        self.ui = uic.loadUi(ui_path, self)

        self.ui.setWindowTitle('Events')

        self.ui.move(self.mpfmon.local_settings.value('windows/events/pos',
                                                   QPoint(500, 200)))
        self.ui.resize(self.mpfmon.local_settings.value('windows/events/size',
                                                     QSize(300, 600)))

        # Disable option "Sort", select first item.
        # TODO: Store and load selected sort index to local_settings
        self.ui.sortComboBox.model().item(0).setEnabled(False)
        self.ui.sortComboBox.setCurrentIndex(1)

    def attach_signals(self):
        assert (self.ui is not None)
        self.ui.filterLineEdit.textChanged.connect(self.filter_text)
        self.ui.sortComboBox.currentIndexChanged.connect(self.change_sort)
        self.ui.clear_button.clicked.connect(self.clear_log)

    def attach_model(self):
        self.model = QStandardItemModel(0, 2)

        self.model.setHeaderData(0, Qt.Orientation.Horizontal, "Event")
        self.model.setHeaderData(1, Qt.Orientation.Horizontal, "Data")
        # self.model.setHeaderData(2, Qt.Orientation.Horizontal, "Time")

        self.filtered_model = QSortFilterProxyModel(self)
        self.filtered_model.setSourceModel(self.model)
        self.filtered_model.setFilterKeyColumn(0)
        self.filtered_model.setDynamicSortFilter(True)

        self.change_sort()  # Default sort

        self.ui.tableView.setModel(self.filtered_model)
        self.ui.tableView.setColumnHidden(2, True)
        self.rootNode = self.model.invisibleRootItem()

    def add_event_to_model(self, event_name, event_type, event_callback,
                             event_kwargs, registered_handlers):
        """Add an event."""
        assert(self.model is not None)
        # remove _from_bcp arg
        event_kwargs.pop('_from_bcp', False)

        name = QStandardItem(event_name)
        kwargs = QStandardItem(str(event_kwargs))
        time_added = QStandardItem(str(self.added_index).zfill(10))
        self.added_index += 1
        self.model.insertRow(0, [name, kwargs, time_added])

    def update_events(self):
        """Update view."""
        self.ui.tableView.resizeColumnToContents(0)
        self.ui.tableView.resizeColumnToContents(1)

        if not self.already_hidden:
            self.ui.tableView.setColumnHidden(2, True)
            self.already_hidden = True

    def filter_text(self, string):
        wc_string = "*" + str(string) + "*"
        self.filtered_model.setFilterWildcard(wc_string)
        self.ui.tableView.resizeColumnToContents(0)
        self.ui.tableView.resizeColumnToContents(1)

    def change_sort(self, index=1):
        # This is a bit sloppy and probably should be reworked.
        if index == 1:  # Received up
            self.filtered_model.sort(2, Qt.SortOrder.DescendingOrder)
        elif index == 2:  # Received down
            self.filtered_model.sort(2, Qt.SortOrder.AscendingOrder)
        elif index == 3:  # Name up
            self.filtered_model.sort(0, Qt.SortOrder.AscendingOrder)
        elif index == 4:  # Name down
            self.filtered_model.sort(0, Qt.SortOrder.DescendingOrder)

    def clear_log(self):
        #clears the log of events
        self.model.clear()

    def closeEvent(self, event):
        self.mpfmon.write_local_settings()
        event.accept()
        self.mpfmon.check_if_quit()



from PyQt6.QtCore import *
from PyQt6.QtGui import *
from PyQt6.QtWidgets import *
from PyQt6 import uic

import os
import time

class ModeWindow(QWidget):

    def __init__(self, mpfmon):
        self.mpfmon = mpfmon
        super().__init__()
        self.ui = None
        self.model = None
        self.draw_ui()
        self.attach_model()
        self.attach_signals()

        self.already_hidden = False
        self.added_index = 0

    def draw_ui(self):
        # Load ui file from ./ui/
        ui_path = os.path.join(os.path.dirname(__file__), "ui", "searchable_table.ui")
        self.ui = uic.loadUi(ui_path, self)
        self.ui.clear_button.hide()

        self.ui.setWindowTitle('Running Modes')

        self.ui.move(self.mpfmon.local_settings.value('windows/modes/pos',
                                                   QPoint(1100, 200)))
        self.ui.resize(self.mpfmon.local_settings.value('windows/modes/size',
                                                     QSize(300, 240)))

        # Fix sort combobox verbiage
        self.ui.sortComboBox.setItemText(1, "Priority ▴")
        self.ui.sortComboBox.setItemText(2, "Priority ▾")


        # Disable option "Sort", select first item.
        # TODO: Store and load selected sort index to local_settings
        self.ui.sortComboBox.model().item(0).setEnabled(False)
        self.ui.sortComboBox.setCurrentIndex(1)

    def attach_signals(self):
        assert (self.ui is not None)
        self.ui.filterLineEdit.textChanged.connect(self.filter_text)
        self.ui.sortComboBox.currentIndexChanged.connect(self.change_sort)

    def attach_model(self):
        self.model = QStandardItemModel(0, 2)

        self.model.setHeaderData(0, Qt.Orientation.Horizontal, "Mode")
        self.model.setHeaderData(1, Qt.Orientation.Horizontal, "Priority")
        # self.model.setHeaderData(2, Qt.Orientation.Horizontal, "Time")

        self.filtered_model = QSortFilterProxyModel(self)
        self.filtered_model.setSourceModel(self.model)
        self.filtered_model.setFilterKeyColumn(0)
        self.filtered_model.setDynamicSortFilter(True)

        self.change_sort()  # Default sort

        self.ui.tableView.setModel(self.filtered_model)
        self.ui.tableView.setColumnHidden(2, True)
        self.rootNode = self.model.invisibleRootItem()

    def process_mode_update(self, running_modes):
        """Update mode list."""
        self.model.clear()

        for mode in running_modes:
            mode_name = QStandardItem(mode[0])
            mode_priority = QStandardItem(str(mode[1]))
            mode_priority_padded = QStandardItem(str(mode[1]).zfill(10))
            self.model.insertRow(0, [mode_name, mode_priority, mode_priority_padded])

        # Reset the headers for the tree. For some reason clear() wipes these too.
        self.model.setHeaderData(0, Qt.Orientation.Horizontal, "Mode")
        self.model.setHeaderData(1, Qt.Orientation.Horizontal, "Priority")

        self.ui.tableView.setColumnHidden(2, True)

    def filter_text(self, string):
        wc_string = "*" + str(string) + "*"
        self.filtered_model.setFilterWildcard(wc_string)
        self.ui.tableView.resizeColumnToContents(0)
        self.ui.tableView.resizeColumnToContents(1)

    def change_sort(self, index=1):
        # This is a bit sloppy and probably should be reworked.
        if index == 1:  # Received up
            self.filtered_model.sort(2, Qt.SortOrder.DescendingOrder)
        elif index == 2:  # Received down
            self.filtered_model.sort(2, Qt.SortOrder.AscendingOrder)
        elif index == 3:  # Name up
            self.filtered_model.sort(0, Qt.SortOrder.AscendingOrder)
        elif index == 4:  # Name down
            self.filtered_model.sort(0, Qt.SortOrder.DescendingOrder)

    def closeEvent(self, event):
        self.mpfmon.write_local_settings()
        event.accept()
        self.mpfmon.check_if_quit()



# mpf-monitor
__version__ = '0.57.2.dev2'
__short_version__ = '0.57'
__bcp_version__ = '1.1'
__config_version__ = '6'

version = f"MPF Monitor v{__version__} (config_version={__config_version__}, BCP v{__bcp_version__})"



import logging
import time
import os

# will change these to specific imports once code is more final
from PyQt6.QtCore import *
from PyQt6.QtGui import *
from PyQt6.QtWidgets import *
from PyQt6 import uic

BRUSH_WHITE = QBrush(QColor(255, 255, 255), Qt.BrushStyle.SolidPattern)
BRUSH_GREEN = QBrush(QColor(0, 255, 0), Qt.BrushStyle.SolidPattern)
BRUSH_BLACK = QBrush(QColor(0, 0, 0), Qt.BrushStyle.SolidPattern)
BRUSH_DARK_PURPLE = QBrush(QColor(128, 0, 255), Qt.BrushStyle.SolidPattern)


class DeviceNode:

    __slots__ = ["_callback", "_name", "_data", "_type", "_brush", "q_name", "q_state", "sub_properties",
                 "sub_properties_appended", "q_time_added", "log"]

    def __init__(self):
        self._callback = None
        self._name = ""
        self._data = {}
        self._type = ""
        self._brush = BRUSH_BLACK

        self.q_name = QStandardItem()
        self.q_state = QStandardItem()
        self.sub_properties = {}
        self.sub_properties_appended = False

        self.q_time_added = QStandardItem()
        self.q_time_added.setData(time.perf_counter(), Qt.ItemDataRole.DisplayRole)

        self.q_name.setDragEnabled(True)
        self.q_state.setData("", Qt.ItemDataRole.DisplayRole)

        self.log = logging.getLogger('Device')

    def setName(self, name):
        self._name = name
        self.q_name.setData(str(self._name), Qt.ItemDataRole.DisplayRole)
        self.log = logging.getLogger('Device {}'.format(self._name))
        self.q_state.emitDataChanged()

    def setData(self, data):
        """Set data of device."""
        if data == self._data:
            # do nothing if data did not change
            return

        if not isinstance(data, dict):
            data = {}

        if self._callback:
            self._callback()

        self._data = data

        state_str = str(list(self._data.values())[0])
        if len(self._data) > 1:
            state_str = state_str + " {…}"
        self.q_state.setData(state_str, Qt.ItemDataRole.DisplayRole)

        for row in self._data:
            if not self.sub_properties_appended:
                q_property = QStandardItem()
                q_value = QStandardItem()
                self.sub_properties.update({row: [q_property, q_value]})
                self.q_name.appendRow(self.sub_properties.get(row))

            self.sub_properties.get(row)[0].setData(str(row), Qt.ItemDataRole.DisplayRole)
            self.sub_properties.get(row)[1].setData(str(self._data.get(row)), Qt.ItemDataRole.DisplayRole)

        self.sub_properties_appended = True
        self.q_state.emitDataChanged()
        self._brush = self._calculate_colored_brush()

    def setType(self, type):
        self._type = type
        self._brush = self._calculate_colored_brush()
        self.q_state.emitDataChanged()

    def get_row(self):
        return [self.q_name, self.q_state, self.q_time_added]

    def data(self):
        return self._data

    def type(self):
        return self._type

    def get_colored_brush(self) -> QBrush:
        """Return colored brush for device."""
        return self._brush

    def _calculate_color_gamma_correction(self, color):
        """Perform gamma correction.

        Feel free to fiddle with these constants until it feels right
        With gamma = 0.5 and constant a = 18, the top 54 values are lost,
        but the bottom 25% feels much more normal.
        """
        gamma = 0.5
        a = 18
        corrected = []

        for value in color:
            if value < 0 or value > 255:
                self.log.warning("Got value %s for brightness which outside the expected range", value)
                value = 0

            value = int(pow(value, gamma) * a)
            if value > 255:
                value = 255
            corrected.append(value)

        return corrected

    def _calculate_colored_brush(self):
        if self._type == 'light':
            color = self.data()['color']
            if color == [0, 0, 0]:
                # shortcut for black
                return BRUSH_BLACK
            color = self._calculate_color_gamma_correction(color)

        elif self._type == 'switch':
            state = self.data()['state']

            if state:
                return BRUSH_GREEN
            else:
                return BRUSH_BLACK

        elif self._type == 'diverter':
            state = self.data()['active']

            if state:
                return BRUSH_DARK_PURPLE
            else:
                return BRUSH_BLACK
        else:
            # Get first parameter and draw as white if it evaluates True
            state = bool(list(self.data().values())[0])
            if state:
                return BRUSH_WHITE
            else:
                return BRUSH_BLACK

        return QBrush(QColor(*color), Qt.BrushStyle.SolidPattern)

    def set_change_callback(self, callback):
        if self._callback:
            # raise AssertionError("Can only have one callback")
            old_callback = self._callback
            self._callback = callback
            return old_callback
        else:
            self._callback = callback
            self.q_state.emitDataChanged()


class DeviceDelegate(QStyledItemDelegate):
    def __init__(self):
        self.size = None
        super().__init__()

    def paint(self, painter, view, index):
        super().paint(painter, view, index)
        color = None
        state = None
        balls = None
        found = False
        text = ''

        # src_index = index.model().mapToSource(index)
        # src_index_model = src_index.model()
        # print(index.data())
        # print(src_index_model.data())
        data = []
        try:
            data = index.model().itemFromIndex(index).data()
            # src_index = index.model().mapToSource(index)
            # data = index.model().data(src_index)
        except:
            pass


        num_circles = 1
        # return

        if index.column() == 0:
            return

        try:
            if 'color' in data:
                color = data['color']
                found = True
        except TypeError:
            return

        try:
            if 'brightness' in data:
                color = [data['brightness']]*3
                found = True
        except TypeError:
            return

        try:
            if 'state' in data:
                text = str(data['state'])
                found = True
        except TypeError:
            return

        try:
            if 'complete' in data:
                state = not data['complete']
                found = True
        except TypeError:
            return

        try:
            if 'enabled' in data:
                state = data['enabled']
                found = True
        except TypeError:
            return

        try:
            if 'balls' in data:
                balls = data['balls']
                found = True
        except TypeError:
            return

        try:
            if 'balls_locked' in data:
                balls = data['balls_locked']
                found = True
        except TypeError:
            return

        try:
            if 'num_balls_requested' in data:
                text += 'Requested: {} '.format(
                    data['num_balls_requested'])
                found = True
        except TypeError:
            return

        try:
            if 'unexpected_balls' in data:
                text += 'Unexpected: {} '.format(
                    data['unexpected_balls'])
                found = True
        except TypeError:
            return

        if not found:
            return

        text += " " + str(data)

        painter.save()

        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setPen(QPen(QColor(100, 100, 100), 1, Qt.PenStyle.SolidLine))

        if color:
            painter.setBrush(QBrush(QColor(*color), Qt.BrushStyle.SolidPattern))
        elif state is True:
            painter.setBrush(QBrush(QColor(0, 255, 0), Qt.BrushStyle.SolidPattern))
        elif state is False:
            painter.setBrush(QBrush(QColor(255, 255, 255), Qt.BrushStyle.SolidPattern))
        elif isinstance(balls, int):
            painter.setBrush(QBrush(QColor(0, 255, 0), Qt.BrushStyle.SolidPattern))
            num_circles = balls

        x_offset = 0
        for _ in range(num_circles):
            painter.drawEllipse(
                view.rect.x() + x_offset, view.rect.y(), 14, 14)

            x_offset += 20

        if text:
            painter.drawText(view.rect.x() + x_offset, view.rect.y() + 12,
                             str(text))
            self.size = QSize(len(text) * 10, 20)

        painter.restore()

    def sizeHint(self, QStyleOptionViewItem, QModelIndex):
        if self.size:
            return self.size
        else:
            # Calling super() here seems to result in a segfault on close sometimes.
            # return super().sizeHint(QStyleOptionViewItem, QModelIndex)
            return QSize(80, 20)


class DeviceWindow(QWidget):

    __slots__ = ["mpfmn", "ui", "model", "log", "already_hidden", "added_index", "device_states",
                 "device_type_widgets", "_debug_enabled"]

    def __init__(self, mpfmon):
        self.mpfmon = mpfmon
        super().__init__()
        self.ui = None
        self.model = None
        self.draw_ui()
        self.attach_model()
        self.attach_signals()

        self.log = logging.getLogger('Core')

        self.already_hidden = False
        self.added_index = 0

        self.device_states = dict()
        self.device_type_widgets = dict()
        self._debug_enabled = self.log.isEnabledFor(logging.DEBUG)

    def draw_ui(self):
        # Load ui file from ./ui/
        ui_path = os.path.join(os.path.dirname(__file__), "ui", "searchable_tree.ui")
        self.ui = uic.loadUi(ui_path, self)

        self.ui.setWindowTitle('Devices')

        self.ui.move(self.mpfmon.local_settings.value('windows/devices/pos',
                                            QPoint(200, 200)))
        self.ui.resize(self.mpfmon.local_settings.value('windows/devices/size',
                                              QSize(300, 600)))

        # Disable option "Sort", select first item.
        # TODO: Store and load selected sort index to local_settings
        self.ui.sortComboBox.model().item(0).setEnabled(False)
        self.ui.sortComboBox.setCurrentIndex(1)
        self.ui.treeView.setAlternatingRowColors(True)

    def attach_signals(self):
        assert (self.ui is not None)
        self.ui.treeView.expanded.connect(self.resize_columns_to_content)
        self.ui.treeView.collapsed.connect(self.resize_columns_to_content)
        self.ui.filterLineEdit.textChanged.connect(self.filter_text)
        self.ui.sortComboBox.currentIndexChanged.connect(self.change_sort)

    def attach_model(self):
        assert (self.ui is not None)
        self.treeview = self.ui.treeView

        self.model = QStandardItemModel()
        self.model.setHorizontalHeaderLabels(["Device", "Data"])

        self.treeview.setDragDropMode(QAbstractItemView.DragDropMode.DragOnly)
        # self.treeview.setItemDelegateForColumn(1, DeviceDelegate())

        # Resizing to contents causes huge performance losses. Only resize when rows expanded or collapsed.
        # self.treeview.header().setSectionResizeMode(QHeaderView.ResizeToContents)

        self.filtered_model = QSortFilterProxyModel(self)
        self.filtered_model.setSourceModel(self.model)
        self.filtered_model.setRecursiveFilteringEnabled(True)
        self.filtered_model.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)

        self.treeview.setModel(self.filtered_model)

    def resize_columns_to_content(self):
        self.ui.treeView.resizeColumnToContents(0)
        self.ui.treeView.resizeColumnToContents(1)

    def process_device_update(self, name, state, changes, type):
        del changes
        if self._debug_enabled:
            self.log.debug("Device Update: %s.%s: %s", type, name, state)

        if type not in self.device_states:
            self.device_states[type] = dict()

            item = QStandardItem(type)
            self.device_type_widgets[type] = item

            self.model.appendRow([item, QStandardItem(), QStandardItem(str(time.perf_counter()))])

        if name not in self.device_states[type]:
            node = DeviceNode()
            node.setName(name)
            node.setData(state)
            node.setType(type)

            self.device_states[type][name] = node
            self.device_type_widgets[type].appendRow(node.get_row())

            self.mpfmon.pf.create_widget_from_config(node, type, name)
        else:
            self.device_states[type][name].setData(state)

        self.ui.treeView.setColumnHidden(2, True)

    def filter_text(self, string):
        wc_string = "*" + str(string) + "*"
        self.filtered_model.setFilterWildcard(wc_string)
        self.ui.treeView.resizeColumnToContents(0)
        self.ui.treeView.resizeColumnToContents(1)

    def change_sort(self, index=1):
        self.model.layoutAboutToBeChanged.emit()
        self.filtered_model.beginResetModel()

        # This is a bit sloppy and probably should be reworked.
        if index == 1:  # Received up
            self.filtered_model.sort(2, Qt.SortOrder.AscendingOrder)
        elif index == 2:  # Received down
            self.filtered_model.sort(2, Qt.SortOrder.DescendingOrder)
        elif index == 3:  # Name up
            self.filtered_model.sort(0, Qt.SortOrder.AscendingOrder)
        elif index == 4:  # Name down
            self.filtered_model.sort(0, Qt.SortOrder.DescendingOrder)

        self.filtered_model.endResetModel()
        self.model.layoutChanged.emit()

    def closeEvent(self, event):
        super().closeEvent(event)
        self.mpfmon.write_local_settings()
        event.accept()
        self.mpfmon.check_if_quit()



import unittest
import threading
import sys
from PyQt6.QtTest import QTest
from PyQt6 import QtCore, QtGui, QtWidgets
from unittest.mock import MagicMock, patch, NonCallableMock
from mpfmonitor.core.devices import *


class TestableDeviceWindowNoGUI(DeviceWindow):
    def __init__(self, mpfmon_mock=None, logger=False):
        if mpfmon_mock is not None:
            self.mpfmon = mpfmon_mock

        if logger:
            self.log = logging.getLogger('Core')

        self.ui = None
        self.model = None

        self.device_states = dict()
        self.device_type_widgets = dict()
        self._debug_enabled = False

class TestDeviceWindowFunctions(unittest.TestCase):
    def setUp(self):
        self.device_window = TestableDeviceWindowNoGUI()
        self.device_window.ui = MagicMock()
        self.device_window.model = MagicMock()
        self.device_window.filtered_model = MagicMock()

    @patch('mpfmonitor.core.devices.QStandardItemModel', autospec=True)
    @patch('mpfmonitor.core.devices.QSortFilterProxyModel', autospec=True)
    def test_attach_model(self, mock_standard_item, mock_proxy_item):
        self.device_window.attach_model()

        self.device_window.model.setHorizontalHeaderLabels.assert_called_once()
        self.device_window.filtered_model.setSourceModel.assert_called_once()
        self.device_window.ui.treeView.setModel.assert_called_once()

    @patch('mpfmonitor.core.devices.QStandardItemModel', autospec=True)
    @patch('mpfmonitor.core.devices.DeviceNode', autospec=True)
    def test_process_device_update(self, node, q_item):
        self.device_window.log = MagicMock()
        self.device_window.mpfmon = MagicMock()

        name = "switch1"
        state =  {'state': 0, 'recycle_jitter_count': 0}
        changes = False
        type = "switch"

        self.device_window.process_device_update(name, state, changes, type)

        self.assertTrue(isinstance(self.device_window.device_states[type], dict))

        self.device_window.model.appendRow.assert_called_once()

        node().setName.assert_called_once_with(name)
        node().setData.assert_called_with(state)
        node().setType.assert_called_once_with(type)

        # self.device_window.device_type_widgets[type].appendRow.assert_called_once_with(node.get_row())

        self.device_window.mpfmon.pf.create_widget_from_config.assert_called_once_with(node(), type, name)

        self.device_window.device_states[type][name].setData.assert_called_with(state)


    def test_filter_text(self):
        string_in = "filter_string_test"
        expected_string_out = "*filter_string_test*"

        self.device_window.filter_text(string=string_in)

        self.device_window.filtered_model.setFilterWildcard.assert_called_once_with(expected_string_out)

    def test_change_sort_default(self):
        self.device_window.change_sort()
        self.device_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.AscendingOrder)

    def test_change_sort_time_down(self):
        self.device_window.change_sort(1)
        self.device_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.AscendingOrder)

    def test_change_sort_time_up(self):
        self.device_window.change_sort(2)
        self.device_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.DescendingOrder)

    def test_change_sort_name_up(self):
        self.device_window.change_sort(3)
        self.device_window.filtered_model.sort.assert_called_once_with(0, Qt.SortOrder.AscendingOrder)

    def test_change_sort_name_down(self):
        self.device_window.change_sort(4)
        self.device_window.filtered_model.sort.assert_called_once_with(0, Qt.SortOrder.DescendingOrder)



if __name__ == '__main__':
    unittest.main()



import logging

# For drag and drop vs click separation
import time

# will change these to specific imports once code is more final
from PyQt6.QtCore import *
from PyQt6.QtGui import *
from PyQt6.QtWidgets import *

from enum import Enum

from mpfmonitor.core.devices import DeviceNode


class Shape(Enum):
    DEFAULT = 0
    SQUARE = 1
    RECTANGLE = 2
    CIRCLE = 3
    TRIANGLE = 4
    ARROW = 5
    FLIPPER = 6


class PfView(QGraphicsView):

    def __init__(self, parent, mpfmon):
        self.mpfmon = mpfmon
        super().__init__(parent)

        self.setWindowTitle("Playfield")
        self.set_inspector_mode_title(inspect=False)

    def resizeEvent(self, event=None):
        self.fitInView(self.mpfmon.pf, Qt.AspectRatioMode.KeepAspectRatio)

    def set_inspector_mode_title(self, inspect=False):
        if inspect:
            self.setWindowTitle('Inspector Enabled - Playfield')
        else:
            self.setWindowTitle("Playfield")

    def closeEvent(self, event):
        self.mpfmon.write_local_settings()
        event.accept()
        self.mpfmon.check_if_quit()


class PfPixmapItem(QGraphicsPixmapItem):

    def __init__(self, image, mpfmon, parent=None):
        super().__init__(image, parent)

        self.mpfmon = mpfmon
        self.setAcceptDrops(True)
        self._height = None
        self._width = None

    def invalidate_size(self):
        self._height = None
        self._width = None

    @property
    def height(self):
        """Return the height of the scene."""
        if self._height is None:
            self._height = self.mpfmon.scene.height()
        return self._height

    @property
    def width(self):
        """Return the width of the scene."""
        if self._width is None:
            self._width = self.mpfmon.scene.width()
        return self._width

    def create_widget_from_config(self, widget, device_type, device_name):
        try:
            x = self.mpfmon.config[device_type][device_name]['x']
            y = self.mpfmon.config[device_type][device_name]['y']
            default_size = self.mpfmon.pf_device_size
            shape_str = self.mpfmon.config[device_type][device_name].get('shape', 'DEFAULT')
            shape = Shape[shape_str]
            rotation = self.mpfmon.config[device_type][device_name].get('rotation', 0)
            size = self.mpfmon.config[device_type][device_name].get('size', default_size)

        except KeyError:
            return

        x *= self.width
        y *= self.height

        self.create_pf_widget(widget, device_type, device_name, x, y,
                              size=size, rotation=rotation, shape=shape, save=False)

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    dragMoveEvent = dragEnterEvent

    def dropEvent(self, event):
        device = event.source().selectedIndexes()[0]
        device_name = device.data()
        device_type = device.parent().data()

        drop_x = event.scenePos().x()
        drop_y = event.scenePos().y()

        try:
            widget = self.mpfmon.device_window.device_states[device_type][device_name]
            self.create_pf_widget(widget, device_type, device_name, drop_x,
                                  drop_y)
        except KeyError:
            self.mpfmon.log.warn("Invalid device dragged.")

    def create_pf_widget(self, widget, device_type, device_name, drop_x,
                         drop_y, size=None, rotation=0, shape=Shape.DEFAULT, save=True):
        w = PfWidget(self.mpfmon, widget, device_type, device_name, drop_x,
                     drop_y, size=size, rotation=rotation, shape_type=shape, save=save)

        self.mpfmon.scene.addItem(w)


class PfWidget(QGraphicsItem):

    def __init__(self, mpfmon, widget, device_type, device_name, x, y,
                 size=None, rotation=0, shape_type=Shape.DEFAULT, save=True):
        super().__init__()

        self.widget = widget    # type: DeviceNode
        self.mpfmon = mpfmon
        self.name = device_name
        self.move_in_progress = True
        self.device_type = device_type
        self.set_size(size=size)
        self.shape_type = shape_type
        self.angle = rotation

        self.setToolTip('{}: {}'.format(self.device_type, self.name))
        self.setAcceptedMouseButtons(Qt.MouseButton.LeftButton | Qt.MouseButton.RightButton)
        self.setPos(x, y)
        self.update_pos(save)
        self.click_start = 0
        self.release_switch = False
        self.pen = QPen(Qt.GlobalColor.white, 3, Qt.PenStyle.SolidLine)

        self.log = logging.getLogger('Core')

        old_widget_exists = widget.set_change_callback(self.notify)

        if old_widget_exists:
            self.log.debug("Previous widget exists.")
            old_widget_exists(destroy=True)


    def boundingRect(self):
        known_points = self.points_for_draw_shape()
        if known_points != None:
            x_options = [sub_list[0] for sub_list in known_points]
            x_min = min(x_options)
            width = max(x_options) - x_min
            y_options = [sub_list[1] for sub_list in known_points]
            y_min = min(y_options)
            height = max(y_options) - y_min
            return QRectF(int(x_min * self.device_size), int(y_min * self.device_size), int(width * self.device_size), int(height * self.device_size))

        else:
            return QRectF(int(self.device_size / -2), int(self.device_size / -2),
                          int(self.device_size), int(self.device_size))

    def set_shape_type(self, shape_type):
        if isinstance(shape_type, Shape):
            self.shape_type = shape_type
        else:
            self.shape_type = Shape.DEFAULT

    def set_rotation(self, angle=0):
        angle = angle % 360
        self.angle = angle

    def set_size(self, size=None):
        if size is None:
            self.size = self.mpfmon.pf_device_size
            self.device_size = self.mpfmon.scene.width() * \
                               self.mpfmon.pf_device_size
        else:
            self.size = size
            self.device_size = self.mpfmon.scene.width() * size

    def resize_to_default(self, force=False):
        device_config = self.mpfmon.config[self.device_type].get(self.name, None)

        if force:
            device_config.pop('size', None) # Delete saved size info, None is incase key doesn't exist (popped twice)

        device_size = device_config.get('size', None)

        if device_size is not None:
            # Do not change the size if it's already set
            pass
        elif device_config is not None:
            self.set_size()

        self.update_pos(save=False)  # Do not save at this point. Let it be saved elsewhere. This reduces writes.

    def draw_shape(self):
        shape_result = self.shape_type

        # Preserve legacy and regular use
        if shape_result == Shape.DEFAULT:
            if self.device_type == 'light':
                shape_result = Shape.CIRCLE

            elif self.device_type == 'switch':
                shape_result = Shape.SQUARE

            elif self.device_type == 'diverter':
                shape_result = Shape.TRIANGLE

            else:  # Draw any other devices as square by default
                shape_result = Shape.SQUARE

        return shape_result


    def paint(self, painter, option, widget=None):
        """Paint this widget to the playfield."""
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setPen(self.pen)
        painter.rotate(self.angle)

        painter.setBrush(self.widget.get_colored_brush())

        draw_shape = self.draw_shape()
        if draw_shape == Shape.CIRCLE:
            painter.drawEllipse(int(self.device_size / -2), int(self.device_size / -2),
                                int(self.device_size), int(self.device_size))
        else:
            shape_points = self.points_for_draw_shape()
            if shape_points != None:
                scaled_points = map(lambda pair: QPoint(int(pair[0] * self.device_size), int(pair[1] * self.device_size)), shape_points)
                painter.drawPolygon(QPolygon(scaled_points))

    def points_for_draw_shape(self):
        draw_shape = self.draw_shape()
        if draw_shape == Shape.CIRCLE:
            return None # Handle circles with drawEllipse instead

        elif draw_shape == Shape.SQUARE:
            return self.square_points()

        elif draw_shape == Shape.RECTANGLE:
            return self.rectangle_points()

        elif draw_shape == Shape.TRIANGLE:
            return self.wide_triangle_points()

        elif draw_shape == Shape.ARROW:
            return self.arrow_points()

        elif draw_shape == Shape.FLIPPER:
            return self.tall_triangle_points()

    def square_points(self):
        return [[-.5, -.5], [.5, -.5], [.5, .5], [-.5, .5]]

    def rectangle_points(self):
        return [[-.2, -.5], [.2, -.5], [.2, .5], [-.2, .5]]

    def wide_triangle_points(self):
        return [[0, -.6], [-.6, .3], [.6, .3]]

    def arrow_points(self):
        return [[0, -.7], [-.4, 0], [-.2, 0], [-.2, .4], [.2, .4], [.2, 0], [.4, 0]]

    def tall_triangle_points(self):
        return [[0, -.3], [-.3, .7], [.3, .7]]

    def notify(self, destroy=False, resize=False):
        self.update()

        if destroy:
            self.destroy()

    def destroy(self):
        self.log.debug("Destroy device: %s", self.name)
        self.mpfmon.scene.removeItem(self)
        self.delete_from_config()

    def mouseMoveEvent(self, event):
        if (self.mpfmon.pf.boundingRect().width() > event.scenePos().x() >
                0) and (self.mpfmon.pf.boundingRect().height() >
                event.scenePos().y() > 0):
            # devices off the pf do weird things at the moment

            if time.time() - self.click_start > .3:
                self.setPos(event.scenePos())
                self.move_in_progress = True

    def mousePressEvent(self, event):
        self.click_start = time.time()

        if self.device_type == 'switch':
            if event.buttons() & Qt.MouseButton.RightButton:
                if not self.get_val_inspector_enabled():
                    self.mpfmon.bcp.send('switch', name=self.name, state=-1)
                    self.release_switch = False
                else:
                    self.send_to_inspector_window()
                    self.log.debug('Switch %s right clicked', self.name)
            elif event.buttons() & Qt.MouseButton.LeftButton:
                if not self.get_val_inspector_enabled():
                    self.mpfmon.bcp.send('switch', name=self.name, state=-1)
                    self.release_switch = True
                else:
                    self.send_to_inspector_window()
                    self.log.debug('Switch %s clicked', self.name)

        else:
            if event.buttons() & Qt.MouseButton.RightButton:
                if self.get_val_inspector_enabled():
                    self.send_to_inspector_window()
                    self.log.debug('%s %s right clicked', self.device_type, self.name)
            elif event.buttons() & Qt.MouseButton.LeftButton:
                if self.get_val_inspector_enabled():
                    self.send_to_inspector_window()
                    self.log.debug('%s %s clicked', self.device_type, self.name)

    def mouseReleaseEvent(self, event):
        if self.move_in_progress and time.time() - self.click_start > .5:
            self.move_in_progress = False
            self.update_pos()

        elif self.release_switch:
            self.mpfmon.bcp.send('switch', name=self.name, state=-1)

        self.click_start = 0

    def update_pos(self, save=True):
        x = self.pos().x() / self.mpfmon.scene.width() if self.mpfmon.scene.width() > 0 else self.pos().x()
        y = self.pos().y() / self.mpfmon.scene.height() if self.mpfmon.scene.height() > 0 else self.pos().y()

        if self.device_type not in self.mpfmon.config:
            self.mpfmon.config[self.device_type] = dict()

        if self.name not in self.mpfmon.config[self.device_type]:
            self.mpfmon.config[self.device_type][self.name] = dict()

        self.mpfmon.config[self.device_type][self.name]['x'] = x
        self.mpfmon.config[self.device_type][self.name]['y'] = y

        # Only save the shape if it is different than the  default
        conf_shape_str = self.mpfmon.config[self.device_type][self.name].get('shape', 'DEFAULT')
        conf_shape = Shape[str(conf_shape_str).upper()]

        if self.shape_type is not conf_shape:
            if self.shape_type is not Shape.DEFAULT:
                self.mpfmon.config[self.device_type][self.name]['shape'] = self.shape_type.name
            else:
                try:
                    self.mpfmon.config[self.device_type][self.name].pop('shape')
                except:
                    pass

        # Only save the rotation if it has been changed
        conf_angle = self.mpfmon.config[self.device_type][self.name].get('angle', -1)

        if self.angle is not conf_angle:
            if self.angle != 0:
                self.mpfmon.config[self.device_type][self.name]['rotation'] = self.angle
            else:
                try:
                    self.mpfmon.config[self.device_type][self.name].pop('rotation')
                except:
                    pass

        # Only save the size if it is different than the top level default
        default_size = self.mpfmon.pf_device_size
        conf_size = self.mpfmon.config[self.device_type][self.name].get('size', default_size)

        if self.size is not conf_size \
                and self.size is not self.mpfmon.pf_device_size:
            self.mpfmon.config[self.device_type][self.name]['size'] = self.size

        if save:
            self.mpfmon.save_config()

    def delete_from_config(self):
        self.mpfmon.config[self.device_type].pop(self.name)
        self.mpfmon.save_config()

    def get_val_inspector_enabled(self):
        return self.mpfmon.inspector_enabled

    def send_to_inspector_window(self):
        self.mpfmon.inspector_window_last_selected_cb(pf_widget=self)



import unittest
import threading
import os
from PyQt6.QtCore import Qt
from PyQt6.QtTest import QTest
from mpfmonitor.core.mpfmon import *


"""class InitMPFMon(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        app = QApplication(sys.argv)
        machine_path = os.path.join(os.getcwd(), "machine_files")
        self.mpfmon_sut = MainWindow(app, machine_path, None, testing=True)
        QTest.qWait(5000)

    def test_case(self):
        self.assertEqual(True, True)

"""

if __name__ == '__main__':
    unittest.main()



"""Starts the MPF Monitor."""

import argparse
import logging
import os
import socket
import sys
import threading
from datetime import datetime
import time

import errno
from mpfmonitor._version import __version__

# import functiontrace
# functiontrace.trace()

class Command(object):

    # pylint: disable-msg=too-many-locals
    def __init__(self, mpf_path, machine_path, args):

        # Need to have these in here because we don't want them to load when
        # the module is loaded as an mpf.command
        from mpf.core.utility_functions import Util

        del mpf_path

        parser = argparse.ArgumentParser(description='Starts the MPF Monitor')

        parser.add_argument("-l",
                            action="store", dest="logfile",
                            metavar='file_name',
                            default=os.path.join("logs", datetime.now().strftime(
                                "%Y-%m-%d-%H-%M-%S-monitor-" +
                                socket.gethostname() +
                                ".log")),
                            help="The name (and path) of the log file")

        parser.add_argument("-c",
                            action="store", dest="configfile",
                            default="monitor", metavar='config_file(s)',
                            help="The name of a config file to load. Note "
                                 "this is a config for the monitor itself, "
                                 "not an MPF config.yaml. Default is "
                                 "monitor.")

        parser.add_argument("-v",
                            action="store_const", dest="loglevel", const=logging.DEBUG,
                            default=logging.INFO, help="Enables verbose logging to the"
                                                       " log file")

        parser.add_argument("-V",
                            action="store_true", dest="consoleloglevel",
                            default=logging.INFO,
                            help="Enables verbose logging to the console. Do NOT on "
                                 "Windows platforms")

        parser.add_argument("-C",
                            action="store", dest="mpfmonconfigfile",
                            default="mpfmonitor.yaml",
                            metavar='config_file',
                            help="The MPF Monitor default config file. "
                                 "Default is <mpf-monitor install "
                                 "folder>/mpfmonitor.yaml")

        parser.add_argument("-ip",
                            action="store", dest="mpfipaddr",
                            help="The MPF IP Address Default is localhost")

        parser.add_argument("-port",
                            action="store", dest="mpfport",
                            help="The MPF Port Default is 5051")

        args = parser.parse_args(args)
        args.configfile = "{}.yaml".format(args.configfile)

        # Configure logging. Creates a logfile and logs to the console.
        # Formatting options are documented here:
        # https://docs.python.org/2.7/library/logging.html#logrecord-attributes

        try:
            os.makedirs(os.path.join(machine_path, 'logs'))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

        logging.basicConfig(level=args.loglevel,
                            format='%(asctime)s : %(levelname)s : %(name)s : '
                                   '%(message)s',
                            filename=os.path.join(machine_path, args.logfile),
                            filemode='w')

        # define a Handler which writes INFO messages or higher to the
        # sys.stderr
        console = logging.StreamHandler()
        console.setLevel(args.consoleloglevel)

        # set a format which is simpler for console use
        formatter = logging.Formatter('%(levelname)s : %(name)s : %(message)s')

        # tell the handler to use this format
        console.setFormatter(formatter)

        # add the handler to the root logger
        logging.getLogger('').addHandler(console)

        from mpfmonitor.core.mpfmon import run

        logging.info("Loading MPF Monitor Version {}".format(__version__))

        thread_stopper = threading.Event()

        try:
            run(machine_path=machine_path,
                thread_stopper=thread_stopper,
                config_file=args.configfile,
                ip_addr=args.mpfipaddr,
                port=args.mpfport)
            logging.info("MPF Monitor run loop ended.")
        except Exception as e:
            logging.exception(str(e))

        logging.info("Stopping child threads... (%s remaining)",
                     len(threading.enumerate()) - 1)

        thread_stopper.set()

        while len(threading.enumerate()) > 1:
            time.sleep(.1)

        logging.info("All child threads stopped.")

        sys.exit()

def get_command():
    return 'monitor', Command



import unittest
from unittest import TestCase
import sys

from mpfmonitor.core.modes import *
from unittest.mock import MagicMock


class TestableModeNoGUI(ModeWindow):
    def __init__(self, mpfmon_mock=None):
        if mpfmon_mock is not None:
            self.mpfmon = mpfmon_mock

        self.ui = None
        self.model = None


class TestModeWindowFunctions(unittest.TestCase):

    def setUp(self):
        self.mode_window = TestableModeNoGUI()

        self.mode_window.ui = MagicMock()
        self.mode_window.model = MagicMock()
        self.mode_window.filtered_model = MagicMock()

    def test_process_mode_update(self):

        modes_in = [
            ["mode1", 100],
            ["mode2", 1000],
            ["mode3", 10000]
        ]

        self.mode_window.process_mode_update(running_modes=modes_in)

        self.mode_window.model.clear.assert_called_once()

        # for mode in modes_in:
        #     mode_name = QStandardItem(mode[0])
        #     mode_priority = QStandardItem(str(mode[1]))
        #     mode_priority_padded = QStandardItem(str(mode[1]).zfill(10))
        #
        #     self.mode_window.model.insertRow.assert_called_with(0, [mode_name, mode_priority, mode_priority_padded])

        # For now, just test it's called as many times as there are modes.
        self.assertEqual(self.mode_window.model.insertRow.call_count, len(modes_in))

    def test_filter_text(self):
        string_in = "filter_string_test"
        expected_string_out = "*filter_string_test*"

        self.mode_window.filter_text(string=string_in)

        self.mode_window.filtered_model.setFilterWildcard.assert_called_once_with(expected_string_out)

    def test_change_sort_default(self):
        self.mode_window.change_sort()
        self.mode_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.DescendingOrder)

    def test_change_sort_time_down(self):
        self.mode_window.change_sort(1)
        self.mode_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.DescendingOrder)

    def test_change_sort_time_up(self):
        self.mode_window.change_sort(2)
        self.mode_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.AscendingOrder)

    def test_change_sort_name_up(self):
        self.mode_window.change_sort(3)
        self.mode_window.filtered_model.sort.assert_called_once_with(0, Qt.SortOrder.AscendingOrder)

    def test_change_sort_name_down(self):
        self.mode_window.change_sort(4)
        self.mode_window.filtered_model.sort.assert_called_once_with(0, Qt.SortOrder.DescendingOrder)


app = QApplication(sys.argv)


class TestModeWindowGUI(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        mock_mpfmon = MagicMock()
        mock_mpfmon.local_settings.value.side_effect = [QPoint(1100, 200), QSize(300, 250)]

        self.mode_window = ModeWindow(mock_mpfmon)

        self.mock_event_kwargs = MagicMock()
        self.mock_event_kwargs.__str__ = MagicMock(return_value='{args}')
        self.mock_event_kwargs.pop.return_value(False)

    def test_model(self):
        self.assertIsNotNone(self.mode_window.model)

    def test_empty_table(self):
        # Reset table model
        self.mode_window.attach_model()

        # Check it's empty
        self.assertEqual(self.mode_window.model.rowCount(), 0)
        self.assertEqual(self.mode_window.filtered_model.rowCount(), 0)

    def test_add_to_table(self):
        # Reset table model
        self.mode_window.attach_model()
        self.assertEqual(self.mode_window.filtered_model.rowCount(), 0)

        modes_in = [
            ["mode1", 100],
            ["mode2", 1000],
            ["mode3", 10000]
        ]

        self.mode_window.process_mode_update(running_modes=modes_in)

        # Check table has 3 rows
        self.assertEqual(self.mode_window.filtered_model.rowCount(), 3)


    def test_sort(self):
        # Reset table model
        self.mode_window.attach_model()
        self.assertEqual(self.mode_window.filtered_model.rowCount(), 0)

        modes_in = [
            ["mode1", 100],
            ["mode2", 1000],
            ["mode3", 10000]
        ]

        self.mode_window.process_mode_update(running_modes=modes_in)

        # Default is Received up
        top_row_text = self.mode_window.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, modes_in[-1][0])

        # Sort Received up
        self.mode_window.ui.sortComboBox.setCurrentIndex(1)
        top_row_text = self.mode_window.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, modes_in[-1][0])

        # Sort Received down
        self.mode_window.ui.sortComboBox.setCurrentIndex(2)
        top_row_text = self.mode_window.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, modes_in[0][0])

        # Sort Name up
        self.mode_window.ui.sortComboBox.setCurrentIndex(3)
        top_row_text = self.mode_window.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, modes_in[0][0])

        # Sort Name down
        self.mode_window.ui.sortComboBox.setCurrentIndex(4)
        top_row_text = self.mode_window.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, modes_in[-1][0])

    def test_filter(self):
        # Reset table model
        self.mode_window.attach_model()
        self.assertEqual(self.mode_window.filtered_model.rowCount(), 0)

        modes_in = [
            ["mode1", 100],
            ["mode32", 1000],
            ["mode3", 10000]
        ]

        self.mode_window.process_mode_update(running_modes=modes_in)

        # Make sure filter is empty and check none are filtered
        self.mode_window.ui.filterLineEdit.setText("")
        self.assertEqual(self.mode_window.filtered_model.rowCount(), len(modes_in))

        # Set the filter to a unique string and check it returns 1 match
        self.mode_window.ui.filterLineEdit.setText(modes_in[0][0])
        self.assertEqual(self.mode_window.filtered_model.rowCount(), 1)

        # Set the filter to a non-unique string and check it returns 2 matches
        self.mode_window.ui.filterLineEdit.setText(modes_in[2][0])
        self.assertEqual(self.mode_window.filtered_model.rowCount(), 2)


if __name__ == '__main__':
    unittest.main()



import unittest
import threading
import sys
from PyQt6.QtTest import QTest
from PyQt6 import QtCore, QtGui, QtWidgets
from unittest.mock import MagicMock
from mpfmonitor.core.events import *


class TestableEventNoGUI(EventWindow):
    def __init__(self, mpfmon_mock=None):
        if mpfmon_mock is not None:
            self.mpfmon = mpfmon_mock

        self.ui = None
        self.model = None

        self.already_hidden = False
        self.added_index = 0


class TestEventWindowFunctions(unittest.TestCase):

    def setUp(self):
        self.event_window = TestableEventNoGUI()

        self.event_window.ui = MagicMock()
        self.event_window.model = MagicMock()
        self.event_window.filtered_model = MagicMock()

        self.mock_event_kwargs = MagicMock()
        self.mock_event_kwargs.__str__ = MagicMock(return_value='{args}')
        self.mock_event_kwargs.pop.return_value(False)

    def test_add_event_to_model(self):
        self.assertEqual(self.event_window.already_hidden, False)

        self.event_window.add_event_to_model("event1", None, None, self.mock_event_kwargs, None)

        self.event_window.model.insertRow.assert_called_once()
        # Disabled by Brian because this assert fails, but I don't know what it's actually testing, feel free to fix & re-enable :)
        # self.assertEqual(self.event_window.already_hidden, True)

    def test_filter_text(self):
        string_in = "filter_string_test"
        expected_string_out = "*filter_string_test*"

        self.event_window.filter_text(string=string_in)

        self.event_window.filtered_model.setFilterWildcard.assert_called_once_with(expected_string_out)

    def test_change_sort_default(self):
        self.event_window.change_sort()
        self.event_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.DescendingOrder)

    def test_change_sort_time_down(self):
        self.event_window.change_sort(1)
        self.event_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.DescendingOrder)

    def test_change_sort_time_up(self):
        self.event_window.change_sort(2)
        self.event_window.filtered_model.sort.assert_called_once_with(2, Qt.SortOrder.AscendingOrder)

    def test_change_sort_name_up(self):
        self.event_window.change_sort(3)
        self.event_window.filtered_model.sort.assert_called_once_with(0, Qt.SortOrder.AscendingOrder)

    def test_change_sort_name_down(self):
        self.event_window.change_sort(4)
        self.event_window.filtered_model.sort.assert_called_once_with(0, Qt.SortOrder.DescendingOrder)


app = QApplication(sys.argv)


class TestEvents(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        mock_mpfmon = MagicMock()
        mock_mpfmon.local_settings.value.side_effect = [QPoint(500, 200), QSize(300, 600)]

        self.eventWindow = EventWindow(mock_mpfmon)

        self.mock_event_kwargs = MagicMock()
        self.mock_event_kwargs.__str__ = MagicMock(return_value='{args}')
        self.mock_event_kwargs.pop.return_value(False)

    def test_model(self):
        self.assertIsNotNone(self.eventWindow.model)

    def test_empty_table(self):
        # Reset table model
        self.eventWindow.attach_model()

        # Check it's empty
        self.assertEqual(self.eventWindow.model.rowCount(), 0)
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), 0)

    def test_add_to_table(self):
        # Reset table model
        self.eventWindow.attach_model()
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), 0)

        self.eventWindow.add_event_to_model("event1", None, None, self.mock_event_kwargs, None)

        # Check table has 1 row
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), 1)

        self.eventWindow.add_event_to_model("event2", None, None, self.mock_event_kwargs, None)
        self.eventWindow.add_event_to_model("event3", None, None, self.mock_event_kwargs, None)

        # Check table has 3 rows
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), 3)

    def test_sort(self):
        # Reset table model
        self.eventWindow.attach_model()
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), 0)

        event_list = ["event_a", "event_b", "event_c"]

        for e in event_list:
            self.eventWindow.add_event_to_model(e, None, None, self.mock_event_kwargs, None)

        # Default is Received up
        top_row_text = self.eventWindow.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, event_list[-1])

        # Sort Received up
        self.eventWindow.ui.sortComboBox.setCurrentIndex(1)
        top_row_text = self.eventWindow.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, event_list[-1])

        # Sort Received down
        self.eventWindow.ui.sortComboBox.setCurrentIndex(2)
        top_row_text = self.eventWindow.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, event_list[0])

        # Sort Name up
        self.eventWindow.ui.sortComboBox.setCurrentIndex(3)
        top_row_text = self.eventWindow.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, event_list[0])

        # Sort Name down
        self.eventWindow.ui.sortComboBox.setCurrentIndex(4)
        top_row_text = self.eventWindow.filtered_model.index(0, 0).data()
        self.assertEqual(top_row_text, event_list[-1])

    def test_filter(self):
        # Reset table model
        self.eventWindow.attach_model()
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), 0)

        event_list = ["abc", "def", "ghi", "ghijkl"]

        for e in event_list:
            self.eventWindow.add_event_to_model(e, None, None, self.mock_event_kwargs, None)

        # Make sure filter is empty and check none are filtered
        self.eventWindow.ui.filterLineEdit.setText("")
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), len(event_list))

        # Set the filter to a unique string and check it returns 1 match
        self.eventWindow.ui.filterLineEdit.setText(event_list[0])
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), 1)

        # Set the filter to a non-unique string and check it returns 2 matches
        self.eventWindow.ui.filterLineEdit.setText(event_list[2])
        self.assertEqual(self.eventWindow.filtered_model.rowCount(), 2)


if __name__ == '__main__':
    unittest.main()



import logging
import queue
import sys
import os
import time

# will change these to specific imports once code is more final
from collections import deque

from PyQt6.QtCore import *
from PyQt6.QtGui import *
from PyQt6.QtWidgets import *

from ruamel import yaml

from mpfmonitor.core.devices import *
from mpfmonitor.core.playfield import *
from mpfmonitor.core.bcp_client import BCPClient
from mpfmonitor.core.events import EventWindow
from mpfmonitor.core.modes import ModeWindow
from mpfmonitor.core.inspector import InspectorWindow
from mpfmonitor.core.variables import VariableWindow

def run(machine_path, thread_stopper, config_file, ip_addr="localhost", port="5051", testing=False):
    app = QApplication(sys.argv)
    MPFMonitor(app, machine_path, thread_stopper, config_file, ip_addr, port, testing=testing)
    app.exec()

class MPFMonitor():
    def __init__(self, app, machine_path, thread_stopper, config_file, ip_addr=None, port=None, parent=None, testing=False):

        # super().__init__(parent)

        self.log = logging.getLogger('Core')

        sys.excepthook = self.except_hook

        self.bcp_client_connected = False
        self.receive_queue = queue.Queue()
        self.sending_queue = queue.Queue()
        self.crash_queue = queue.Queue()
        self.thread_stopper = thread_stopper
        self.machine_path = machine_path
        self.app = app
        self.config = None
        self.layout = None
        self.mpf_ip_addr = ip_addr
        self.mpf_port = port
        self.config_file = os.path.join(self.machine_path, "monitor",
                                        config_file)
        self.playfield_image_file = os.path.join(self.machine_path,
                                                 "monitor", "playfield.jpg")
        self.settings_file = os.path.join(self.machine_path,
                                                 "monitor", "settings.ini")

        QSettings.setDefaultFormat(QSettings.Format.IniFormat)
        self.local_settings = QSettings(self.settings_file, QSettings.Format.IniFormat)

        self.load_config()

        self.device_window = DeviceWindow(self)

        self.pf_device_size = self.config.get("device_size", .02)
        if not isinstance(self.pf_device_size, float):  # Protect against corrupted device size
            self.pf_device_size = .02

        #Command line takes priority over settings file. If command line is None, then read the settings file.

        #if mpf_ip_address is not in the settings file, then localhost will be used
        if (self.mpf_ip_addr == None):
            self.mpf_ip_addr = self.local_settings.value("settings/mpf-ip-address", "localhost")

        #if mpf_port is not in the settings file, then 5051 will be used
        if (self.mpf_port == None):
            self.mpf_port = self.local_settings.value("settings/mpf-port", "5051")

        self.bcp = BCPClient(self, self.receive_queue,
                             self.sending_queue, self.mpf_ip_addr, self.mpf_port,
                             simulate=testing, cache=False)

        self.tick_timer = QTimer(self.device_window)
        self.tick_timer.setInterval(20)
        self.tick_timer.timeout.connect(self.tick)
        self.tick_timer.start()

        self.toggle_pf_window_action = QAction('&Playfield', self.device_window,
                                        statusTip='Show the playfield window',
                                        triggered=self.toggle_pf_window)
        self.toggle_pf_window_action.setCheckable(True)

        self.toggle_device_window_action = QAction('&Devices', self.device_window,
                                        statusTip='Show the device window',
                                        triggered=self.toggle_device_window)
        self.toggle_device_window_action.setCheckable(True)

        self.toggle_event_window_action = QAction('&Events', self.device_window,
                                        statusTip='Show the events window',
                                        triggered=self.toggle_event_window)
        self.toggle_event_window_action.setCheckable(True)

        self.toggle_mode_window_action = QAction('&Modes', self.device_window,
                                        statusTip='Show the mode window',
                                        triggered=self.toggle_mode_window)
        self.toggle_mode_window_action.setCheckable(True)

        self.toggle_variables_window_action = QAction('&Variables', self.device_window,
                                        statusTip='Show the variables window',
                                        triggered=self.toggle_variables_window)
        self.toggle_variables_window_action.setCheckable(True)

        self.scene = QGraphicsScene()

        self.pf = PfPixmapItem(QPixmap(self.playfield_image_file), self)
        self.scene.addItem(self.pf)

        self.view = PfView(self.scene, self)

        self.view.move(self.local_settings.value('windows/pf/pos',
                                                 QPoint(800, 200)))
        self.view.resize(self.local_settings.value('windows/pf/size',
                                                   QSize(300, 600)))

        self.event_window = EventWindow(self)

        self.variables_window = VariableWindow(self)
        # self.variables_window.show()

        self.mode_window = ModeWindow(self)

        if self.get_local_settings_bool('windows/pf/visible'):
            self.toggle_pf_window()

        if self.get_local_settings_bool('windows/events/visible'):
            self.toggle_event_window()

        if self.get_local_settings_bool('windows/devices/visible'):
            self.toggle_device_window()

        if self.get_local_settings_bool('windows/modes/visible'):
            self.toggle_mode_window()

        if self.get_local_settings_bool('windows/variables/visible'):
            self.toggle_variables_window()

        self.exit_on_close = False

        if self.get_local_settings_bool('settings/exit-on-close'):
            self.toggle_exit_on_close()

        self.inspector_enabled = False

        self.inspector_window = InspectorWindow(self)
        self.inspector_window.show()
        self.inspector_window.register_last_selected_cb()

        self.inspector_window.register_set_inspector_val_cb(self.set_inspector_mode)

        self.menu_bar = QMenuBar()
        self.view_menu = self.menu_bar.addMenu("&View")
        self.view_menu.addAction(self.toggle_pf_window_action)
        self.view_menu.addAction(self.toggle_device_window_action)
        self.view_menu.addAction(self.toggle_event_window_action)
        self.view_menu.addAction(self.toggle_mode_window_action)
        self.view_menu.addAction(self.toggle_variables_window_action)

    def toggle_pf_window(self):
        if self.view.isVisible():
            self.view.hide()
            self.toggle_pf_window_action.setChecked(False)
        else:
            self.view.show()
            self.toggle_pf_window_action.setChecked(True)

    def toggle_device_window(self):
        if self.device_window.isVisible():
            self.device_window.hide()
            self.toggle_device_window_action.setChecked(False)
        else:
            self.device_window.show()
            self.toggle_device_window_action.setChecked(True)

    def toggle_event_window(self):
        if self.event_window.isVisible():
            self.event_window.hide()
            self.toggle_event_window_action.setChecked(False)
        else:
            self.event_window.show()
            self.toggle_event_window_action.setChecked(True)

    def toggle_mode_window(self):
        if self.mode_window.isVisible():
            self.mode_window.hide()
            self.toggle_mode_window_action.setChecked(False)
        else:
            self.mode_window.show()
            self.toggle_mode_window_action.setChecked(True)

    def toggle_variables_window(self):
        if self.variables_window.isVisible():
            self.variables_window.hide()
            self.toggle_variables_window_action.setChecked(False)
        else:
            self.variables_window.show()
            self.toggle_variables_window_action.setChecked(True)

    def toggle_exit_on_close(self):
        if self.exit_on_close:
            self.exit_on_close = False
        else:
            self.exit_on_close = True

    def toggle_sort_by_time(self):
        if self.sort_by_time:
            self.sort_by_time = False
        else:
            self.sort_by_time = True

    def except_hook(self, cls, exception, traceback):
        sys.__excepthook__(cls, exception, traceback)
        self.app.exit()

    def reset_connection(self):
        self.start_time = 0
        self.event_window.model.clear()
        self.mode_window.model.clear()

    def eventFilter(self, source, event):
        try:
            if source is self.playfield and event.type() == QEvent.Resize:
                self.playfield.setPixmap(self.playfield_image.scaled(
                    self.playfield.size(), Qt.AspectRatioMode.KeepAspectRatio,
                    Qt.TransformationMode.SmoothTransformation))
                self.pf.invalidate_size()
        except AttributeError:
            pass

        return super().eventFilter(source, event)

    def tick(self):
        """
        Called every 20 mSec
        Check the queue to see if BCP has any messages to process.
        If any devices have updated, refresh the model data.
        """
        # get the complete queue
        with self.receive_queue.mutex:
            local_queue = self.receive_queue.queue
            self.receive_queue.queue = deque()

        added_events = False
        for cmd, kwargs in local_queue:
            if cmd == 'device':
                self.device_window.process_device_update(**kwargs)
            elif cmd == 'monitored_event':
                self.event_window.add_event_to_model(**kwargs)
                added_events = True
            elif cmd in ('mode_start', 'mode_stop', 'mode_list'):
                if 'running_modes' not in kwargs:
                    # ignore mode_start/stop on newer MPF versions
                    continue
                self.mode_window.process_mode_update(kwargs['running_modes'])
            elif cmd == 'reset':
                self.reset_connection()
                self.bcp.send("reset_complete")
            elif cmd == 'player_variable':
                self.variables_window.update_variable("player", kwargs["name"], kwargs["value"])
            elif cmd == 'machine_variable':
                self.variables_window.update_variable("machine", kwargs["name"], kwargs["value"])

        if added_events:
            self.event_window.update_events()

    def about(self):
        QMessageBox.about(self, "About MPF Monitor",
                "This is the MPF Monitor")

    def load_config(self):
        try:
            _yaml = yaml.YAML(typ='safe')
            with open(self.config_file, 'r') as f:
                self.config = _yaml.load(f)
        except FileNotFoundError:
                self.config = dict()

    def save_config(self):
        self.log.debug("Saving config to disk")
        with open(self.config_file, 'w') as f:
            _yaml = yaml.YAML(typ='safe')
            _yaml.default_flow_style = False
            _yaml.dump(self.config, f)

    def closeEvent(self, event):
        self.write_local_settings()
        event.accept()
        self.check_if_quit()

    def check_if_quit(self):
        if self.exit_on_close:
            self.log.info("Quitting due to quit on close")
            QCoreApplication.exit(0)

    def write_window_settings(self, window_name, window):
        settings = {
            'pos': window.pos(),
            'size': window.size(),
            'visible': window.isVisible()
        }
        for line in settings.keys():
            setting_name = 'windows/' + window_name + '/' + line
            self.local_settings.setValue(setting_name, settings.get(line))

    def get_local_settings_bool(self, setting):
        return "true" == str(self.local_settings.value(setting, False)).lower()

    def write_local_settings(self):

        monitor_windows = {
            'devices': self.device_window,
            'pf': self.view,
            'modes': self.mode_window,
            'events': self.event_window,
            'inspector': self.inspector_window,
            'variables': self.variables_window,
        }

        for window in monitor_windows.keys():
            self.write_window_settings(window, monitor_windows.get(window))

        self.local_settings.setValue('settings/exit-on-close', self.exit_on_close)

        self.local_settings.sync()

    def set_inspector_mode(self, enabled=False):
        self.inspector_enabled = enabled
        self.view.set_inspector_mode_title(inspect=enabled)
